dofile("sha256.lua")
dofile("pnotesocket.lua")

local function worker(sock,key,tag)
   local err = sock:GetSocketError()
   if err then error("connection failed, is your Interwebs down? "..err) end
   pnotesocket.send(sock,"NONCE PLZ")
   local msg = assert(pnotesocket.receive(sock))
   assert(msg:sub(1,7) == "NONCE: ")
   local nonce = msg:sub(8,-1)
   pnotesocket.send(sock,"I AM "..SHA256(key..nonce).." LOOKING FOR "..tag)
   msg = assert(pnotesocket.receive(sock))
   if msg == "TAG GONE" then
      print("\127ff0000pnotes: Server claims the pnote tag \""..tag.."\" no longer exists!")
      return false
   elseif msg == "AUTH FAIL" then
      print("\127ff0000pnotes: Server claims the password for tag \""..tag.."\" was wrong!\nPlease get the right password and use /pnotes to enter it.")
      return false
   elseif msg:sub(1,9) ~= "AUTH OK: " then
      print("\127ff0000pnotes: Server gave us a weird response for tag \""..tag.."\"")
      return false
   end
   local stuff = {name=msg:sub(10,-1)}
   repeat
      msg = assert(pnotesocket.receive(sock))
      local a,b,text = msg:match("^System ([0-9]+) Sector ([0-9]+): (.*)$")
      if a then
	 a,b = tonumber(a),tonumber(b)
	 stuff[a] = stuff[a] or {}
	 stuff[a][b] = text:gsub("\n+$","")
      else
	 local a,text = msg:match("^System ([0-9]+): (.*)$")
	 if a then
	    a = tonumber(a)
	    stuff[a] = stuff[a] or {}
	    stuff[a].name = text:gsub("\n+$","")
	 elseif msg:match("Finished") then
	    --print("\12700ff00pnotes: Finished for tag "..tag)
	    return stuff
	 else
	    print("\127ff0000pnotes: NO idea what the server said to us for tag "..tag)
	    return false
	 end
      end
   until false
   print("\127ff0000pnotes: NOTREACHED reached! for tag "..tag)
   return false
end

local function slim(text)
   text = "\n"..text.."\n"
   text = text:gsub("\n%[.*$","\n")
   text = text:gsub("\n[ \t]+\n","\n\n")
   text = text:gsub("^\n+","")
   text = text:gsub("\n+$","")
   if text == "" then return "",0
   else return text,1 end
end

local notes = {}
local connections = {}
local timer
local fail_attempts
local function please_stop_crashing_vo()
   if fail_attempts <= 0 then
      for tag,conn in pairs(connections) do
	 connections[tag] = nil
      end
      return
   else
      fail_attempts = fail_attempts - 1
   end
   for tag,conn in pairs(connections) do
      if conn.go then
	 local s,e = coroutine.resume(conn.co, conn.sock, conn.key, conn.tag)
	 if not s then
	    print("\127ff0000pnotes:"..tag..": "..e)
	    connections[tag] = nil
	 elseif e == "read" or e == "write" then
	    -- nothing
	 elseif e == false then
	    connections[tag] = nil
	 elseif type(e) == "table" then
	    notes[tag] = e
	    connections[tag] = nil
	 else
	    print("\127ff0000pnotes:"..tag..": internal protocol violation")
	    connections[tag] = nil
	 end
      end
   end
end
local function handler()
   for k,v in pairs(connections) do
      timer:SetTimeout(100, handler)
      return please_stop_crashing_vo()
   end
   timer = nil
   --print("\127ffff00All done: "..spickle(notes))
   local t = {}
   for k,v in pairs(notes) do
      t[#t+1] = v
   end
   table.sort(t,function(a,b) return a.name < b.name end)
   -- trim existing [sections] from notes
   for sysid,systable in pairs(SystemNotes) do
      local tn = 0
      local n
      systable.name,n = slim(systable.name)
      tn = tn + n
      for sectorid,text in pairs(systable) do
	 if type(sectorid) == "number" then
	    text,n = slim(text)
	    if n > 0 then
	       tn = tn + n
	       systable[sectorid] = text
	    else
	       systable[sectorid] = nil
	    end
	 end
      end
      if tn == 0 then SystemNotes[sysid] = nil end
   end
   -- now, add each P-note section
   for n=1,#t do
      local pnote = t[n]
      for sysid,psystable in pairs(pnote) do
	 if type(sysid) == "number" then
	    if not SystemNotes[sysid] then SystemNotes[sysid] = {name=""} end
	    local systable = assert(SystemNotes[sysid])
	    if psystable.name then
	       if #systable.name > 0 then
		  systable.name = systable.name.."\n\n["..pnote.name.."]\n"..psystable.name
	       else
		  systable.name = "["..pnote.name.."]\n"..psystable.name
	       end
	    end
	    for sectorid,ptext in pairs(psystable) do
	       if type(sectorid) == "number" then
		  if systable[sectorid] then
		     systable[sectorid] = systable[sectorid].."\n\n["..pnote.name.."]\n"..ptext
		  else
		     systable[sectorid] = "["..pnote.name.."]\n"..ptext
		  end
	       end
	    end
	 end
      end
   end
end

local function reload_infos()
   fail_attempts = 600
   local keystring = gkini.ReadString("P-Notes","keys","open=open") or ""
   for tag,key in keystring:gmatch("([-A-Za-z0-9 ]+)=([^,]+)") do
      local co = coroutine.create(worker)
      local sock = assert(TCPSocket())
      local conn = {co=co,sock=sock,key=key,tag=tag}
      function conn.handler()
	 sock:SetReadHandler(nil)
	 sock:SetWriteHandler(nil)
	 local s,e = coroutine.resume(conn.co, conn.sock, conn.key, conn.tag)
	 if not s then
	    print("\127ff0000pnotes:"..tag..": "..e)
	    connections[tag] = nil
	 elseif e == "read" then
	    conn.go = true
	    --sock:SetReadHandler(conn.handler)
	 elseif e == "write" then
	    conn.go = true
	    --sock:SetWriteHandler(conn.handler)
	 elseif e == false then
	    connections[tag] = nil
	 elseif type(e) == "table" then
	    notes[tag] = e
	    connections[tag] = nil
	 else
	    print("\127ff0000pnotes:"..tag..": internal protocol violation")
	    connections[tag] = nil
	 end
      end
      sock:SetWriteHandler(conn.handler)
      connections[tag] = conn
      assert(sock:Connect("p-notes.tejat.net",14505))
   end
   timer = Timer()
   timer:SetTimeout(100, handler)
end

declare("pnotes_reload_infos",reload_infos)

local function ingame()
   local keystring = gkini.ReadString("P-Notes","keys","open=open")
   if not keystring or keystring == "" then
      print("\127ffffffYou have P-Notes installed, but no keys are present. Please use /pnotes to configure it.")
   else
      --print("\127ffffffRemember, use /pnotes to add, remove, and modify P-Notes access keys.")
   end
   reload_infos()
end

RegisterEvent(ingame, "PLAYER_ENTERED_GAME")

dofile("gui.lua")
