declare("pnotesocket",{})
local buffers = {}
local function construct_message(message)
   local msg = ("%010i%s"):format(#message,message)
   return SHA256(msg)..msg
end
function pnotesocket.send(sock, message)
   message = construct_message(message)
   local pos = 1
   repeat
      local sent = sock:Send(message:sub(pos,-1))
      if sent then
         pos = pos + sent
      else
	 local err = sock:GetSocketError()
	 if err then error("socket error: "..err) end
	 coroutine.yield("write")
      end
   until pos > #message
end
local function pread(sock, amount)
   local text = buffers[sock] or ""
   while #text < amount do
      local a,e = sock:Recv()
      if a and #a > 0 then
	 text = text .. a
      elseif e and e ~= 0 then
	 error("errno "..e)
      else
	 coroutine.yield("read")
      end
   end
   if #text == amount then
      buffers[sock] = nil
   else
      buffers[sock] = text:sub(amount+1,-1)
   end
   return text:sub(1,amount)
end
function pnotesocket.receive(sock)
   local sum = pread(sock,64)
   if sum:match("[^0-9a-f]") then
      error("invalid checksum")
   end
   local len = pread(sock,10)
   if len:match("[^0-9]") then
      error("invalid message length")
   end
   local message = pread(sock,tonumber(len))
   if SHA256(len..message) ~= sum then
      error("checksum mismatch")
   end
   return message
end
