-- modified from the SHA-256 code from SubCritical and luasupersolra
-- luasupersolra is under an unknown license, but SubCritical is LGPL
-- the terms of LGPL 2.1 apply to this file

local function rotateleft(op,amt)
   return bitlib.rshift(op,32-amt) + bitlib.lshift(op,amt)
end

local k = {
        0x428a2f98, 0x71374491, 0xb5c0fbcf, 0xe9b5dba5, 0x3956c25b, 0x59f111f1, 0x923f82a4, 0xab1c5ed5,
        0xd807aa98, 0x12835b01, 0x243185be, 0x550c7dc3, 0x72be5d74, 0x80deb1fe, 0x9bdc06a7, 0xc19bf174,
        0xe49b69c1, 0xefbe4786, 0x0fc19dc6, 0x240ca1cc, 0x2de92c6f, 0x4a7484aa, 0x5cb0a9dc, 0x76f988da,
        0x983e5152, 0xa831c66d, 0xb00327c8, 0xbf597fc7, 0xc6e00bf3, 0xd5a79147, 0x06ca6351, 0x14292967,
        0x27b70a85, 0x2e1b2138, 0x4d2c6dfc, 0x53380d13, 0x650a7354, 0x766a0abb, 0x81c2c92e, 0x92722c85,
        0xa2bfe8a1, 0xa81a664b, 0xc24b8b70, 0xc76c51a3, 0xd192e819, 0xd6990624, 0xf40e3585, 0x106aa070,
        0x19a4c116, 0x1e376c08, 0x2748774c, 0x34b0bcb5, 0x391c0cb3, 0x4ed8aa4a, 0x5b9cca4f, 0x682e6ff3,
        0x748f82ee, 0x78a5636f, 0x84c87814, 0x8cc70208, 0x90befffa, 0xa4506ceb, 0xbef9a3f7, 0xc67178f2
}

-- modified slightly to not support messages larger than 65527 bytes, which
-- would be useless to us and slightly simplifies the code
declare("SHA256", function(message)
   assert(#message < 65528, "message too long")
   local spiffymessage = message .. "\128"
   if #spiffymessage % 64 ~= 56 then
      if #spiffymessage % 64 > 56 then
	 spiffymessage = spiffymessage .. "\0\0\0\0\0\0\0\0"
	 -- band-aid
      end
      spiffymessage = spiffymessage .. ("\0"):rep(56 - #spiffymessage % 64)
   end
   spiffymessage = spiffymessage ..
      ("\0\0\0\0\0\0" .. string.char(math.floor(#message / 256)) .. string.char(#message % 256))
   local rem = #spiffymessage / 64 - 1
   local w = {}
   local h0,h1,h2,h3,h4,h5,h6,h7 = 0x6a09e667, 0xbb67ae85, 0x3c6ef372, 0xa54ff53a, 0x510e527f, 0x9b05688c, 0x1f83d9ab, 0x5be0cd19
   local a,b,c,d,e,f,g,h
   for n=0,rem do
      for m=1,16 do
	 w[m] = string.byte(spiffymessage:sub(n*64+m*4-3,n*64+m*4-3)) * 16777216 +
	    string.byte(spiffymessage:sub(n*64+m*4-2,n*64+m*4-2)) * 65536 +
	    string.byte(spiffymessage:sub(n*64+m*4-1,n*64+m*4-1)) * 256 +
	    string.byte(spiffymessage:sub(n*64+m*4,n*64+m*4))
      end
      for m=17,64 do
	 w[m] = (w[m-16] +
	    bitlib.bxor(bitlib.bxor(rotateleft(w[m-15],25),
				    rotateleft(w[m-15],14)),
			bitlib.rshift(w[m-15],3))
	    + w[m-7]
	    + bitlib.bxor(bitlib.bxor(rotateleft(w[m-2],15),
				      rotateleft(w[m-2],13)),
			  bitlib.rshift(w[m-2],10))) % 4294967296
      end
      a,b,c,d,e,f,g,h = h0,h1,h2,h3,h4,h5,h6,h7
      for m=1,64 do
	 local s0,maj,s1,ch,t1
	 s0 = bitlib.bxor(bitlib.bxor(rotateleft(a,30),rotateleft(a,19)),rotateleft(a,10))
	 maj = bitlib.bxor(bitlib.bxor(bitlib.band(a,b),bitlib.band(a,c)),bitlib.band(b,c))
	 s1 = bitlib.bxor(bitlib.bxor(rotateleft(e,26),rotateleft(e,21)),rotateleft(e,7))
	 ch = bitlib.bxor(bitlib.band(e,f),bitlib.band(bitlib.bnot(e),g))
	 t1 = (h + s1 + ch + k[m] + w[m]) % 4294967296
	 h,g,f,e,d,c,b,a = g,f,e,(d+t1)%4294967296,c,b,a,(t1+s0+maj)%4294967296
	 h0,h1,h2,h3,h4,h5,h6,h7 =
	    (h0+a)%4294967296,
	    (h1+b)%4294967296,
	    (h2+c)%4294967296,
	    (h3+d)%4294967296,
	    (h4+e)%4294967296,
	    (h5+f)%4294967296,
	    (h6+g)%4294967296,
	    (h7+h)%4294967296
      end
   end
   return ("%08x%08x%08x%08x%08x%08x%08x%08x"):format(h0,h1,h2,h3,h4,h5,h6,h7)
end)
